
/*
 *  Michael Palin: "Is it raining ?"
 */

#include "standard.h"
#include "hamdr_cfg.h"

#define WX_INTERVAL 10   /* seconds between samples */

#define MINUTES(n) (n * 60 / WX_INTERVAL)

/* number of anemopulses received, x- and y-wise, signed */

STATIC struct windcomp {
	long x, y;
} veclog[MINUTES(1)], *veclogger = veclog;

/* miles/hour in speedlog[] */

STATIC word speedlog[MINUTES(5)], *speedlogger = speedlog;

extern dword ow_accvalue(word rank);
extern dword read_pulsacc(stamp *time);
extern byte *add_cse_spd(byte *, word course, word knots);

static word atan2(long, long);

/* 5 minute peak gust */

word
windgust()
{
	word max = 0, *p = speedlog; /* need not traverse in chronological order */

	do {
		if (max < *p)
			max = *p;
	} while (++p < &speedlog[__SIZEOF(speedlog)]);

	return (max);
}

/* 1 minute sustained, recent ones from speedlog. #define "sustained" XXX */

word
windspeed()
{
	word sum = 0;           /* "sustained" == average (?) */
	word *p = speedlogger;  /* one past last */

	byte n = MINUTES(1);    /* assume less than 256 samples in 1 minute */

	do {
		if (--p < speedlog)
			p = &speedlog[__SIZEOF(speedlog) - 1]; /* wrap to last */
		sum += *p;
	} while (--n);

	return (sum / MINUTES(1));
}

/* 1 minute average */

word
winddir()
{
	long ew = 0, ns = 0;
	struct windcomp *p = veclog; /* no need for chronological traversal */

	/* cumulate north/south and east/west totals */

	do {
		ew += p->x;
		ns += p->y;
	} while (++p < &veclog[__SIZEOF(veclog)]);

	return (atan2(ew, ns));
}

/*
 *  sensor RH = (Vad/Vdd - 0.16) / 0.0062;
 *  temp compensation = 1.0546 - 0.00216 * celsius;
 *
 *  assume Vdd is 10000h and Vad (AN7) is thus already the ratio.
 *  typical Vad 0.8v ... 3.9v is 41 ... 199 decimal (MSByte).
 */
word
relative_humidity()
{
	word  rh   = read_AN7();   /* unsigned 0...FFC0h from HIH-3610 sensor */
	short temp = ow_celsius(); /* signed celsius from 1st DS1820 */

	if (temp <  -50)       /* huh ? */
		temp =  -50;
	else
	if (temp > +100)
		temp = +100;

	if (rh >= 10486)
		rh -= 10486;
	else
		rh = 0;

	rh = muldiv(rh, 16129, 34557 - 70 * temp);

	rh = (rh + 100) / 200;      /* round and remove 100/0.5 scale */

	return (rh > 100 ? 100 : rh);
}

/* relative sine component of wind from direction 'dir', magnitude 'magn' */

static long
jorgio(word deg, word magn)
{
	bool neg;
	long x;

	if (magn == 0)
		magn = 1;    /* avoid loss of information */

	if (neg = (deg >= 180))
		deg -= 180;

	x = EMUL(magn, eight_bit_sin_from_abs_degrees(deg));

	return (neg ? lnegate(x) : x);
}

/* 'magn' is "amount of wind received lately" */

static void
log_wind_sample(word deg, word magn, word mph)
{
	long x, y;              /* wind components */

	x = jorgio(deg, magn);  /* sine */

	deg += 90;              /* offset to get the ... */
	if (deg >= 360)
		deg -= 360;

	y = jorgio(deg, magn);  /* ... cosine */

	hold_txproc();

	veclogger->x = x;
	veclogger->y = y;

	if (++veclogger >= &veclog[__SIZEOF(veclog)])
		veclogger = veclog;

	*speedlogger = mph;

	if (++speedlogger >= &speedlog[__SIZEOF(speedlog)])
		speedlogger = speedlog;

	okay_txproc();
}

/* aaargh */

static word
mph_to_knots(word mph)
{
	return (muldiv(mph, 577, 664));
}

static word
mph_to_kmh(word mph)
{
	return (muldiv(mph, 1619, 1006));
}

/*
 *  a common transducer has 10 Hz per 1 m/s wind
 *  our reports must use knots for windspeed and mph for gusts.
 *  using mph internally.
 *  assuming 10 second nominal interval:
 *
 *  50 mph, 22.35 m/s, 2235 ticks / 10 seconds
 *  2231 * 2237 / 10'000 / 10 = 49
 *  2277 * 2237 / 10'000 / 10 = 51
 *  1 mph, 0.4470 m/s, 44 ticks / 10 seconds
 *  44 * 2237 / 10'000 / 10 = 0
 *  45 * 2237 / 10'000 / 10 = 1
 *  sigh. good enough
 */

/*
 *  put new pulses into *count
 *  return elapsed msec (nominally 10'000).
 */
static word
anemometer(word *count)
{
	static stamp mark;
	static dword old;

	stamp now;
	dword n;

	n = read_pulsacc(&now) - old;
	old += n;
	if (n > 65535)
		n = 65535;     /* absurd speed, or time problem */
	*count = n;        /* new pulses */

	n = now - mark;
	mark = now;
	if (n > 65535)
		n = 0;         /* eek? too long since last. skip this sample */
	return (n);        /* elapsed msec */
}

/* de-Gray XXX hw-specific */

static word
vane_to_deg[] = {
	  0,   6,  17,  11,  39,  34,  23,  28,
	 84,  79,  68,  73,  45,  51,  62,  56,
	174, 169, 158, 163, 135, 141, 152, 146,
	 90,  96, 107, 101, 129, 124, 113, 118,
	354, 349, 338, 343, 315, 321, 332, 326,
	270, 276, 287, 281, 309, 304, 293, 298,
	180, 186, 197, 191, 219, 214, 203, 208,
	264, 259, 248, 253, 225, 231, 242, 236
};

/*
 *  called every WX_INTERVAL seconds
 *
 *  determine current wind speed and direction,
 *  log the values for averaging. maybe print to serial.
 *
 *  The "/ 10" in mph calculation below, is part of anemoscale,
 *  _not_ related to the (10 second ?) interval.
 *  Extra digit of precision this way, 2237 / 10 or 223.7.
 *  $ units
 *  You have: 0.1 meter / millisecond
 *  You want: mile / hour
 *   * 223.69363
 */
static void
weather_station()
{
	word msec, n, dir, mph;

	msec = anemometer(&n);
	if (!msec)
		return;                 /* eek? too soon (or too long since) */

	dir = (vane_to_deg[windvane()] + cfg.vaneoffset) % 360u;
	mph = muldiv(n, cfg.anemoscale, msec) / 10;

	log_wind_sample(dir, n, mph);

	if (cfg.wx_dump) {
		print_string("wx ");

		print_short(dir);              /* not the averaged direction */
		print_string("  ");
		print_short(mph_to_kmh(mph));
		print_string(" km/h  ");
		print_signed_short(ow_celsius());
		print_string("C  ");
		print_short(relative_humidity());
		print_string("%");

		eol();
	}
}

/* partial arctangent, [0...45] degrees. b >= a */

static word
atan2tab(dword a, dword b)
{
	static byte tab[] = {
		 0,  1,  2,  3,  4,
		 5,  6,  7,  8,  9,
		10, 11, 12, 13, 14,
		15, 16, 17, 18, 19,
		20, 21, 22, 23, 24, /* XXX unity below [25], could save some bytes */
		24, 25, 26, 27, 28,
		29, 29, 30, 31, 32,
		32, 33, 34, 35, 35,
		36, 37, 37, 38, 39,
		39, 40, 41, 41, 42,
		42, 43, 43, 44, 44, /* 55 slots --- max error 1 degree */
		45, ~0              /* 45 degree case and a padding byte */
	};

	if (b == 0)
		return (0);

	/* only divide-by-word available, so prescale both */

	while (b >= 0x10000) {
		LSRx(&a, 1);
		LSRx(&b, 1);
	}
	/* muldiv() suffices now; a < b and b < 0x10000 */

	return (tab[muldiv(55, a, b)]);
}

word
atan2(long ew, long ns)
{
	word deg, q = 0;

	/* locate quadrant and get absolute values */

	if (ew < 0) {
		ew = lnegate(ew);
		q |= 2;
	}
	if (ns < 0) {
		ns = lnegate(ns);
		q |= 1;
	}

	/* 45 degree mirrored */

	if (ns > ew)
		deg = atan2tab(ew, ns);
	else
		deg = 90 - atan2tab(ns, ew);

	switch (q) {
	case 0:
		return (deg);
	case 1:
		return (180 - deg);
	case 2:
		return (360 - deg);
	case 3:
		return (180 + deg);
	}
}

/* APRS weather report */

byte *
encode_weather_ascii(byte *p)
{
	/* CSE/SPD - wind direction, degrees / speed, knots, 1 minute sustained */

	p = add_cse_spd(p, winddir(), mph_to_knots(windspeed()));

	/* 5 minute gust peak, mph - g999 */
	/* temperature, fahrenheit - t999 or t-99 if negative */

	p[0] = 'g';
	sprint_03u(&p[1], windgust());
	p[4] = 't';
	sprint_03d_signed(&p[5], ow_fahrenheit());
	p += 8;

	/* end of fixed part. any other data ? */

	/* relative humidity, percent - h99, h00 if 100% */

	sprint_03u(p, relative_humidity());
	p[0] = 'h';
	p += 3;

	/* barometric pressure, .1 millibars - b99999, that is 5 digits */

	return (p);
}

void
weather_run()
{
	static stamp wx_mark;

	if (time_past(wx_mark)) {
		future_sec(&wx_mark, WX_INTERVAL);

		ow_round();
		weather_station();
	}
}

